/*
 Copyright (c) 2020 gematik GmbH

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 */
package de.gematik.ti.sgd.domain.sgdCommunication;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.JsonSerializer;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.databind.annotation.JsonSerialize;
import de.gematik.ti.sgd.crypto.SgdCryptoUtilities;
import de.gematik.ti.sgd.domain.json.SgdResponseStatus;
import de.gematik.ti.sgd.exceptions.server.SgdServerException;

import java.io.IOException;
import java.security.KeyPair;
import java.security.PublicKey;

@JsonSerialize(using = EncryptedMessage.EncryptedMessageSerializer.class)
public class EncryptedMessage {
    private static final String CURVE_NAME = "brainpoolP256r1";
    private final String content;
    @JsonIgnore
    private final String[] encryptedMessageParts;
    @JsonIgnore
    private final KeyPair messageEncryptionKeyPair;

    public EncryptedMessage(PublicKey receiverPublic, String messagePayload) {
        final SgdCryptoUtilities sgdCryptoUtilities = SgdCryptoUtilities.getInstance();
        this.messageEncryptionKeyPair = sgdCryptoUtilities.generateEccKeyPair();

        this.content = CURVE_NAME + " "
                + sgdCryptoUtilities.extractPublicPointCoordinatesFromPublicKey(receiverPublic)
                + " "
                + sgdCryptoUtilities.extractPublicPointCoordinatesFromPublicKey(messageEncryptionKeyPair.getPublic())
                + " "
                + SgdCryptoUtilities.getInstance().convertBytesToB64String(
                sgdCryptoUtilities.deriveKeyAndEncrypt(messageEncryptionKeyPair.getPrivate(), receiverPublic, messagePayload.getBytes()));

        encryptedMessageParts = content.split(" ");
    }

    public EncryptedMessage(String content) {
        this.content = content;
        this.messageEncryptionKeyPair = null;
        encryptedMessageParts = content.split(" ");
    }

    public String getContent() {
        return content;
    }

    public byte[] decryptMessage(KeyPair keyPair) {
        return decryptMessage(keyPair, null);
    }

    public byte[] decryptMessage(KeyPair keyPairA, KeyPair keyPairB) {
        PublicKey senderPublicKey = getSenderPublicKey();
        byte[] cipher = getCiphertext();
        try {
            return SgdCryptoUtilities.getInstance().deriveKeyAndDecrypt(keyPairA.getPrivate(), senderPublicKey, cipher);
        } catch (RuntimeException ex) {
            try {
                return SgdCryptoUtilities.getInstance().deriveKeyAndDecrypt(keyPairB.getPrivate(), senderPublicKey, cipher);
            } catch (RuntimeException e) {
                throw new SgdServerException("Error during Ecies decryption", ex, SgdResponseStatus.DECRYPTION_FAIL);
            }
        }
    }

    public SgdMessage decryptToSgdMessage(KeyPair keyPair) {
        return decryptToSgdMessage(keyPair, null);
    }

    public SgdMessage decryptToSgdMessage(KeyPair keyPairA, KeyPair keyPairB) {
        String payload = new String(decryptMessage(keyPairA, keyPairB));
        return SgdMessage.parseSgdMessageFromString(payload);
    }

    public byte[] getCiphertext() {
        return SgdCryptoUtilities.getInstance().convertB64StringToBytes(encryptedMessageParts[5]);
    }

    public PublicKey getSenderPublicKey() {
        return SgdCryptoUtilities.getInstance().extractPublicKeyFromPublicPointCoordinates(encryptedMessageParts[3], encryptedMessageParts[4], encryptedMessageParts[0]);
    }

    public PublicKey getReceiverPublicKey() {
        return SgdCryptoUtilities.getInstance().extractPublicKeyFromPublicPointCoordinates(encryptedMessageParts[1], encryptedMessageParts[2], encryptedMessageParts[0]);
    }

    public KeyPair getMessageEncryptionKeyPair() {
        return messageEncryptionKeyPair;
    }

    public static class EncryptedMessageSerializer extends JsonSerializer<EncryptedMessage> {
        @Override
        public void serialize(EncryptedMessage value, JsonGenerator jsonGenerator, SerializerProvider serializerProvider) throws IOException {
            jsonGenerator.writeString(value.getContent());
        }
    }
}
