/*
 Copyright (c) 2020 gematik GmbH

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 */
package de.gematik.ti.sgd.components;

import de.gematik.ti.sgd.SgdTestCryptoUtilities;
import de.gematik.ti.sgd.crypto.SgdCryptoUtilities;
import de.gematik.ti.sgd.domain.json.SgdRequest;
import de.gematik.ti.sgd.domain.json.SgdResponse;
import de.gematik.ti.sgd.domain.json.SgdResponseStatus;
import de.gematik.ti.sgd.exceptions.server.SgdServerException;
import org.junit.Before;
import org.junit.Test;
import org.springframework.test.util.ReflectionTestUtils;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.security.KeyPair;
import java.security.cert.X509Certificate;
import java.util.List;
import java.util.stream.Collectors;

import static de.gematik.ti.sgd.components.SgdClientUtils.extractPublicKeysFromListOfCertificateFilenames;
import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;

public class SgdClientTest {
    {
        SgdCryptoUtilities.setInstance(new SgdTestCryptoUtilities());
    }

    private static List<String> SIGNER_LIST;
    private final String P12_1_FILENAME = "src/test/resources/keys/egk/109500969_X114428530_c.ch.enc-ecc.p12";
    private final String P12_2_FILENAME = "src/test/resources/keys/egk/109500969_X114428530_c.ch.aut-ecc.p12";
    private final String P12_3_FILENAME = "src/test/resources/keys/egk/109500969_X114428531_c.ch.aut-ecc.p12";
    private final String P12_4_FILENAME = "src/test/resources/keys/egk/109500969_X114428533_c.ch.aut-ecc.p12";

    private SgdClient client;
    private SchluesselGenerierungsDienst sgd1;
    private SchluesselGenerierungsDienst sgd2;
    private Sgd1RetryServerWrapper sgd1RetryServerWrapper = new Sgd1RetryServerWrapper();

    private X509Certificate sgd1_Certificate;
    private X509Certificate sgd2_Certificate;
    private KeyPair sgd1_Key;
    private KeyPair sgd2_Key;
    private KeyPair sgd1EphemeralKeyPair;
    private KeyPair sgd2EphemeralKeyPair;
    private KeyPair clientEphemeralKeyPair;
    private KeyPair clientCertificateKeyPair;
    private X509Certificate clientCertificate;

    @Before
    public void setup() throws IOException {
        SIGNER_LIST = Files.list(Path.of("src/test/resources/keys/ca/"))
                .map(path -> path.toString())
                .collect(Collectors.toList());

        sgd1_Certificate = SgdCryptoUtilities.getInstance().loadX509CertificateFromP12File("src/test/resources/keys/sgd/c.sgd1.hsm-aut-ecc.p12", "00");
        sgd2_Certificate = SgdCryptoUtilities.getInstance().loadX509CertificateFromP12File("src/test/resources/keys/sgd/c.sgd2.hsm-aut-ecc.p12", "00");
        sgd1_Key = SgdCryptoUtilities.getInstance().loadKeypairFromP12File("src/test/resources/keys/sgd/c.sgd1.hsm-aut-ecc.p12", "00", "00");
        sgd2_Key = SgdCryptoUtilities.getInstance().loadKeypairFromP12File("src/test/resources/keys/sgd/c.sgd2.hsm-aut-ecc.p12", "00", "00");
        sgd1EphemeralKeyPair = SgdCryptoUtilities.getInstance().loadKeypairFromP12File(P12_1_FILENAME, "00", "00");
        sgd2EphemeralKeyPair = SgdCryptoUtilities.getInstance().loadKeypairFromP12File(P12_4_FILENAME, "00", "00");
        clientEphemeralKeyPair = SgdCryptoUtilities.getInstance().loadKeypairFromP12File(P12_2_FILENAME, "00", "00");
        clientCertificateKeyPair = SgdCryptoUtilities.getInstance().loadKeypairFromP12File(P12_3_FILENAME, "00", "00");
        clientCertificate = SgdCryptoUtilities.getInstance().loadX509CertificateFromP12File(P12_3_FILENAME, "00");

        SgdClientKeyDerivationMachine keyDerivationMachine = new SgdClientKeyDerivationMachine(clientCertificate, clientCertificateKeyPair, extractPublicKeysFromListOfCertificateFilenames(SIGNER_LIST));
        ReflectionTestUtils.setField(keyDerivationMachine, "ephemeralEciesKeyPair", clientEphemeralKeyPair);

        sgd1 = new SchluesselGenerierungsDienst(sgd1_Certificate, sgd1_Key.getPrivate(), sgd1EphemeralKeyPair);
        sgd2 = new SchluesselGenerierungsDienst(sgd2_Certificate, sgd2_Key.getPrivate(), sgd2EphemeralKeyPair);
        ReflectionTestUtils.setField(sgd1, "possibleCertificateSigners", extractPublicKeysFromListOfCertificateFilenames(SIGNER_LIST));
        ReflectionTestUtils.setField(sgd2, "possibleCertificateSigners", extractPublicKeysFromListOfCertificateFilenames(SIGNER_LIST));

        client = new SgdClient(clientCertificate, clientCertificateKeyPair, SIGNER_LIST,
                sgd1RetryServerWrapper, request -> sgd2.receive(request));
    }

    @Test
    public void testBasicDerivation() {
        assertThat(client.getSgd1AesGcmKey())
                .isNull();

        client.performInitialSgdKeyDerivation();

        assertThat(client.getSgd1AesGcmKey())
                .isNotNull();
    }

    @Test
    public void fourTimesRetry_shouldStillGiveSuccess() {
        ReflectionTestUtils.setField(sgd1RetryServerWrapper, "numberOfRetryResponsesLeft", 4);

        assertThat(client.getSgd1AesGcmKey())
                .isNull();

        client.performInitialSgdKeyDerivation();

        assertThat(client.getSgd1AesGcmKey())
                .isNotNull();
    }

    @Test
    public void fiveTimesRetry_shouldGiveError() {
        ReflectionTestUtils.setField(sgd1RetryServerWrapper, "numberOfRetryResponsesLeft", 5);

        assertThatThrownBy(() -> client.performInitialSgdKeyDerivation())
                .isInstanceOf(SgdServerException.class);
    }

    private class Sgd1RetryServerWrapper implements Sgd1ServerWrapper {
        public int numberOfRetryResponsesLeft = 0;

        @Override
        public SgdResponse sgd1Endpoint(SgdRequest request) {
            if (numberOfRetryResponsesLeft-- > 0) {
                throw new SgdServerException("go fish", SgdResponseStatus.RESTART_PROTOCOL);
            }
            return sgd1.receive(request);
        }
    }
}