/*
 Copyright (c) 2020 gematik GmbH

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 */
package de.gematik.ti.sgd.domain;

import de.gematik.ti.sgd.SgdTestCryptoUtilities;
import de.gematik.ti.sgd.crypto.SgdCryptoUtilities;
import de.gematik.ti.sgd.domain.util.PHRUtils;
import de.gematik.ws.fa.phr.v1.PHRKeyType;
import de.gematik.ws.fd.phrs.authorizationservice.v1.EncryptedKeyContainerType;
import org.apache.commons.codec.binary.Base64;
import org.apache.commons.io.FileUtils;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.junit.BeforeClass;
import org.junit.Ignore;
import org.junit.Test;

import javax.xml.bind.DatatypeConverter;
import javax.xml.namespace.QName;
import java.io.File;
import java.io.IOException;
import java.security.Security;

import static de.gematik.ti.sgd.domain.util.PHRUtils.marshal;
import static de.gematik.ti.sgd.domain.util.PHRUtils.unmarshal;
import static org.assertj.core.api.Assertions.assertThat;

public class RecordAndContextKeyTest {
    public static final QName PHR_KEY_ELEMENT_NAME = new QName("http://ws.gematik.de/fa/phr/v1.1", "PHRKey", "epa");
    public static final QName ENCRYPTED_KEY_CONTAINER_ELEMENT_NAME = new QName("http://ws.gematik.de/fd/phrs/AuthorizationService/v1.1", "EncryptedKeyContainer", "phrs");

    {
        SgdCryptoUtilities.setInstance(new SgdTestCryptoUtilities());
    }

    @BeforeClass
    public static void registerBouncyCastle() {
        Security.addProvider(new BouncyCastleProvider());
    }

    @Test
    public void loadUnencryptedPhrRecord_ShouldContainCorrectKvnr() throws IOException {
        RecordAndContextKey phr = readPhrFromFile();
        assertThat(phr.getKvnr()).isEqualTo("A123456789");
    }

    @Test
    public void loadUnencryptedPhrRecord_ShouldContainCorrectContextKey() throws IOException {
        RecordAndContextKey phr = readPhrFromFile();

        assertThat(phr.getContextKeyBytes())
                .containsExactly(Base64.decodeBase64("S2V5Mi0yNTZCaXQtQUVTLUdDTS1iYTA5ODc2NTQzMjE="));
    }

    @Test
    public void loadUnencryptedPhrRecord_ShouldContainCorrectRecordKey() throws IOException {
        RecordAndContextKey phr = readPhrFromFile();

        assertThat(phr.getRecordKeyBytes())
                .containsExactly(Base64.decodeBase64("S2V5MS0yNTZCaXQtQUVTLUdDTS0xMjM0NTY3ODkwYWI="));
    }

    @Test
    public void encryptAndDecryptContainer_ContentShouldBeUnchanged() throws Exception {
        RecordAndContextKey phr = readPhrFromFile();

        final byte[] sgd1KeyBytes = DatatypeConverter.parseHexBinary(FileUtils.readFileToString(new File("src/test/resources/phr/sgd1.key")));
        final byte[] sgd1Vector = FileUtils.readFileToString(new File("src/test/resources/phr/sgd1.vector")).getBytes();

        final EncryptedKeyContainerType encryptedKeyContainer = phr.encipherOnceToContainer(sgd1KeyBytes, sgd1Vector);
        final PHRKeyType decipheredPhrRecord = PHRUtils.decipherContainerToPHRKeyType(encryptedKeyContainer, sgd1KeyBytes);

        assertThat(decipheredPhrRecord.getContextKey().getValue())
                .containsExactly(phr.getContextKeyBytes());
        assertThat(decipheredPhrRecord.getRecordKey().getValue())
                .containsExactly(phr.getRecordKeyBytes());
    }

    @Test
    public void encryptAndDecryptContainerTwice_ContentShouldBeUnchanged() throws Exception {
        RecordAndContextKey phr = readPhrFromFile();

        final byte[] sgd1KeyBytes = DatatypeConverter.parseHexBinary(FileUtils.readFileToString(new File("src/test/resources/phr/sgd1.key")));
        final byte[] sgd1Vector = FileUtils.readFileToString(new File("src/test/resources/phr/sgd1.vector")).getBytes();

        final EncryptedKeyContainerType encryptedKeyContainer = phr.encipherOnceToContainer(sgd1KeyBytes, sgd1Vector);
        final PHRKeyType decipheredPhrRecord = PHRUtils.decipherContainerToPHRKeyType(encryptedKeyContainer, sgd1KeyBytes);

        assertThat(decipheredPhrRecord.getContextKey().getValue())
                .containsExactly(phr.getContextKeyBytes());
        assertThat(decipheredPhrRecord.getRecordKey().getValue())
                .containsExactly(phr.getRecordKeyBytes());
    }

    @Test
    public void performDoubleEncryptionAndDoubleDecryption_ContentShouldBeUnchanged() throws Exception {
        RecordAndContextKey phr = readPhrFromFile();

        final byte[] sgd1KeyBytes = DatatypeConverter.parseHexBinary(FileUtils.readFileToString(new File("src/test/resources/phr/sgd1.key")));
        final String sgd1Vector = FileUtils.readFileToString(new File("src/test/resources/phr/sgd1.vector"));

        final byte[] sgd2KeyBytes = DatatypeConverter.parseHexBinary(FileUtils.readFileToString(new File("src/test/resources/phr/sgd2.key")));
        final String sgd2Vector = FileUtils.readFileToString(new File("src/test/resources/phr/sgd2.vector"));

        final EncryptedKeyContainerType encryptedKeyContainer = phr.encipherTwiceToContainer(sgd1KeyBytes, sgd1Vector, sgd2KeyBytes, sgd2Vector);
        final PHRKeyType decipheredPhrRecord = PHRUtils.decipherTwiceFromContainer(encryptedKeyContainer, sgd1KeyBytes, sgd2KeyBytes,
                PHRUtils.buildOuterAesGcmAdFromOuterAssociatedData(encryptedKeyContainer.getAssociatedData()));

        assertThat(decipheredPhrRecord.getContextKey().getValue())
                .containsExactly(phr.getContextKeyBytes());
        assertThat(decipheredPhrRecord.getRecordKey().getValue())
                .containsExactly(phr.getRecordKeyBytes());
    }

    @Test
    @Ignore("Spezifikation: Das chiffrierte XML ist nicht korrekt!")
    public void decryptSingleStepSpecificationExample_ResultShouldBeReadableAndContainExpectedValues() throws Exception {
        RecordAndContextKey phr = readPhrFromFile();
        final byte[] sgd1KeyBytes = DatatypeConverter.parseHexBinary(FileUtils.readFileToString(new File("src/test/resources/phr/sgd1.key")));
        final String phrString = FileUtils.readFileToString(new File("src/test/resources/phr/phr_nach_erster_verschluesselung.xml"));
        EncryptedKeyContainerType encryptedKeyContainer = unmarshal(phrString.getBytes(), EncryptedKeyContainerType.class);

        PHRKeyType decipheredExample = PHRUtils.decipherContainerToPHRKeyType(encryptedKeyContainer, sgd1KeyBytes);

        assertThat(decipheredExample.getRecordKey().getValue()).containsExactly(phr.getRecordKeyBytes());
        assertThat(decipheredExample.getContextKey().getValue()).containsExactly(phr.getContextKeyBytes());
    }

    @Test
    @Ignore("Test dient zur Unterstützung der Spec")
    public void writeAllTheCorrectXmls() throws Exception {
        RecordAndContextKey phr = readPhrFromFile();

        final byte[] sgd1KeyBytes = DatatypeConverter.parseHexBinary(FileUtils.readFileToString(new File("src/test/resources/phr/sgd1.key")));
        final String sgd1Vector = FileUtils.readFileToString(new File("src/test/resources/phr/sgd1.vector"));

        final byte[] sgd2KeyBytes = DatatypeConverter.parseHexBinary(FileUtils.readFileToString(new File("src/test/resources/phr/sgd2.key")));
        final String sgd2Vector = FileUtils.readFileToString(new File("src/test/resources/phr/sgd2.vector"));

        final EncryptedKeyContainerType doubleEncryptedContainer = phr.encipherTwiceToContainer(sgd1KeyBytes, sgd1Vector, sgd2KeyBytes, sgd2Vector);
        final EncryptedKeyContainerType singleEncryptedContainer = phr.encipherOnceToContainer(sgd1KeyBytes, sgd1Vector.getBytes());

        FileUtils.writeByteArrayToFile(new File("./phr_nach_erster_verschluesselung.xml"), marshal(singleEncryptedContainer, ENCRYPTED_KEY_CONTAINER_ELEMENT_NAME));
        FileUtils.writeByteArrayToFile(new File("./phr_nach_zweiter_verschluesselung.xml"), marshal(doubleEncryptedContainer, ENCRYPTED_KEY_CONTAINER_ELEMENT_NAME));
        FileUtils.writeByteArrayToFile(new File("./phr_unverschlüsselt.xml"), marshal(phr.packToPhrObject(), PHR_KEY_ELEMENT_NAME));
    }

    private RecordAndContextKey readPhrFromFile() throws IOException {
        return PHRUtils.readRecordAndContextKeyFromPhrKeyXml(FileUtils.readFileToByteArray(new File("src/test/resources/phr/phr_unverschluesselt.xml")));
    }
}
